package com.kakao.s2;

import android.app.Application;
import android.content.Context;

import com.google.android.gms.ads.identifier.AdvertisingIdClient;
import com.google.android.gms.common.GooglePlayServicesNotAvailableException;
import com.google.android.gms.common.GooglePlayServicesRepairableException;
import com.kakao.s2.persistence.EventsDataSource;
import com.kakao.network.callback.ResponseCallback;
import com.kakao.network.response.ResponseBody;
import com.kakao.network.tasks.KakaoResultTask;
import com.kakao.network.tasks.KakaoTaskQueue;
import com.kakao.util.exception.KakaoException;
import com.kakao.util.helper.CommonProtocol;
import com.kakao.util.helper.SystemInfo;
import com.kakao.util.helper.Utility;
import com.kakao.util.helper.log.Logger;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

/**
 * 유저의 특정 이벤트를 수집하는 API
 * @author kevin.kang
 * Created by kevin.kang on 2016. 8. 22..
 */
public class S2Service {

    private static S2Service mInstance;

    private Context context;

    /**
     * 유저가 보내는 이벤트들을 batching하기 위해 메모리에 저장하고 있는 이벤트 리스트.
     */
    private List<Event> events;

    private int mBatchTime = 10; // default batch time = 10
    private Timer timer;

    private ResponseCallback<EventsLogResponse> mPublishCallback;
    private ResponseCallback<Integer> mAddEventCallback;

    private EventsDataSource eventsDataSource;
    private final Object INSTANCE_LOCK = new Object();

    /**
     * S2Service 싱글턴 인스턴스를 반환하는 메소드.
     * @return S2Service 싱글턴 인스턴스
     */
    public static S2Service getInstance() {
        if (mInstance == null) {
            throw new KakaoException(KakaoException.ErrorType.MISS_CONFIGURATION, "You should call S2Service.init() first.");
        }
        return mInstance;
    }

    S2Service(Context context, EventsDataSource dataSource) {
        this.context = context;
        this.eventsDataSource = dataSource;
        eventsDataSource.open();
        events = Collections.synchronizedList(eventsDataSource.getAllEvents());
    }

    /**
     * S2 이벤트 API를 사용할 수 있도록 다양한 초기화를 한다.
     *
     * 1. ActivityLifecycleCallback을 등록하여 앱이 foreground에 있을 때만 주기적 배칭을 하도록 한다.
     * 2. 인자로 받은 Application의 getApplicationContext()를 호출하여 추후에 사용할 수 있도록 저장한다.
     * 3. SystemInfo.initialize()를 호출하여 API call을 할 때 헤더 값이 잘 채워지도록 한다.
     *
     * @param application 커스텀 Application 클래스에서 S2Service.init()을 통해 직접 넘겨준 application 객체
     */
    public synchronized static void init(Application application, int batchTime) {
        if (mInstance != null) {
            throw new AlreadyInitializedException();
        }
        if (application == null) {
            throw new KakaoException("You should provide Application instance to initialize S2 Service.");
        }

        SystemInfo.initialize(application.getApplicationContext());
        EventsDataSource eventsDataSource = new EventsDataSource(application.getApplicationContext());

        mInstance = new S2Service(application.getApplicationContext(), eventsDataSource);

        if (batchTime >= 0) {
            mInstance.setBatchTime(batchTime);
        }

        application.registerActivityLifecycleCallbacks(new ActivityLifecycleHandler());
    }

    /**
     * 주기적 배치 타임을 조정 (초 단위)
     * @param time 배치된 이벤트들을 자동으로 S2로 전송할 주기 (초 단위)
     */
    public void setBatchTime(int time) {
        mBatchTime = time;
    }

    /**
     * 이벤트를 보낼 때 default로 사용될 ResponseCallback
     * @param callback  ResponseCallback
     */
    public void setPublishCallback(final ResponseCallback<EventsLogResponse> callback) {
        mPublishCallback = callback;
    }

    /**
     * 이벤트를 batch에 추가할 때 default로 사용될 ResponseCallback
     * @param callback  ResponseCallback
     */
    public void setAddEventCallback(final ResponseCallback<Integer> callback) {
        mAddEventCallback = callback;
    }


    /**
     * 하나의 이벤트를 S2 서버로 전송한다.
     * @param event event 전송할 이벤트
     */
    public void publishEvent(final Event event) {
        publishEventInternal(event, false, mPublishCallback);
    }

    /**
     * 하나의 이벤트를 S2 서버로 전송하고 파라미터로 제공된 callback을 실행한다.
     * @param event 전송할 이벤트
     * @param callback 이벤트 전송 요청 결과에 대한 handler. Null일 경우 mCallback이 불려진다.
     */
    public void publishEvent(final Event event, final ResponseCallback<EventsLogResponse> callback) {
        publishEventInternal(event, false, callback);
    }

    /**
     * 하나의 이벤트를 ADID 정보를 채워 S2 서버로 전송한다.
     * @param event event 전송할 이벤트
     */
    public void publishAdidEvent(final Event event) {
        publishEventInternal(event, true, mPublishCallback);
    }

    /**
     * 하나의 이벤트를 ADID 정보를 채워 S2 서버로 전송하고 파라미터로 제공된 callback을 실행한다.
     * @param event 전송할 이벤트
     * @param callback 이벤트 전송 요청 결과에 대한 handler. Null일 경우 mCallback이 불려진다.
     */
    public void publishAdidEvent(final Event event, final ResponseCallback<EventsLogResponse> callback) {
        publishEventInternal(event, true, callback);
    }

    private void publishEventInternal(final Event event, final boolean fillAdid, final ResponseCallback<EventsLogResponse> callback) {
        KakaoTaskQueue.getInstance().addTask(new KakaoResultTask<EventsLogResponse>(callback) {
            @Override
            public EventsLogResponse call() throws Exception {
                if (event == null) {
                    throw new KakaoException(KakaoException.ErrorType.ILLEGAL_ARGUMENT, "Event cannot be null");
                }
                if (fillAdid) {
                    fillAdidForEvent(event);
                }
                if (event.getFrom() == null) {
                    throw new KakaoException(KakaoException.ErrorType.ILLEGAL_ARGUMENT, "Event's from field cannot be null.");
                }
                EventsLogResponse response = logEvents(event, new LinkedList<Event>());
                if (response == null) {
                    throw new KakaoException(KakaoException.ErrorType.UNSPECIFIED_ERROR, "There was no response from the request.");
                }
                return response;
            }
        });
    }

    /**
     * 하나의 이벤트를 이벤트 batch에 추가한다.
     * @param event batch할 이벤트
     */
    public void addEvent(final Event event) {
        addEventInternal(event, false, mAddEventCallback);
    }

    /**
     * 하나의 이벤트를 이벤트 batch에 추가하고 파라미터로 제공된 callback을 실행한다.
     * @param event batch할 이벤트
     * @param callback 이벤트 배치 요청 결과에 대한 handler
     */
    public void addEvent(final Event event, final ResponseCallback<Integer> callback) {
        addEventInternal(event, false, callback);

    }

    /**
     * 하나의 이벤트를 ADID 정보를 채워 이벤트 batch에 추가한다.
     * @param event batch할 이벤트
     */
    public void addAdidEvent(final Event event) {
        addEventInternal(event, true, mAddEventCallback);
    }

    /**
     * 하나의 이벤트를 ADID 정보를 채워 이벤트 batch에 추가하고 파라미터로 제공된 callback을 실행한다.
     * @param event batch할 이벤트
     */
    public void addAdidEvent(final Event event, final ResponseCallback<Integer> callback) {
        addEventInternal(event, true, callback);
    }

    private void addEventInternal(final Event event, final boolean fillAdid, final ResponseCallback<Integer> callback) {
        KakaoTaskQueue.getInstance().addTask(new KakaoResultTask<Integer>(callback) {
            @Override
            public Integer call() throws Exception {
                if (event == null) {
                    throw new KakaoException(KakaoException.ErrorType.ILLEGAL_ARGUMENT, "Event cannot be null");
                }
                if (fillAdid) {
                    fillAdidForEvent(event);
                }
                if (event.getFrom() == null) {
                    throw new KakaoException(KakaoException.ErrorType.ILLEGAL_ARGUMENT, "Event's from field cannot be null.");
                }
                synchronized (INSTANCE_LOCK) {
                    events.add(event);
                    eventsDataSource.insertEvent(event);
                    return events.size();
                }
            }
        });
    }


    /**
     * batch된 이벤트들을 수동으로 S2 서버로 전송한다.
     */
    public void publishEvents() {
        publishEvents(mPublishCallback);
    }

    /**
     * batch된 이벤트들을 수동으로 S2 서버로 전송하고 파라미터로 제공된 callback을 실행한다.
     * @param callback 이벤트 전송 요청 결과에 대한 handler
     */
    public void publishEvents(final ResponseCallback<EventsLogResponse> callback) {
        final List<Event> copiedEvents;
        synchronized (INSTANCE_LOCK) {
            if (events.size() == 0) {
                return;
            }
            copiedEvents = new ArrayList<Event>(events);
            events.clear();
            eventsDataSource.deleteAllEvents();
        }
        KakaoTaskQueue.getInstance().addTask(new KakaoResultTask<EventsLogResponse>(callback) {
            @Override
            public EventsLogResponse call() throws Exception {
                return logEvents(null, copiedEvents);
            }
        });
    }


    private EventsLogResponse logEvents(final Event rootEvent, final List<Event> leafEvents) throws EventsLogResponse.EventsLogErrorResponseException, ResponseBody.ResponseBodyException, IOException, KakaoException {
        String appKey = Utility.getMetadata(context, CommonProtocol.APP_KEY_PROPERTY);
        if (Utility.isNullOrEmpty(appKey)) {
            throw new KakaoException(KakaoException.ErrorType.MISS_CONFIGURATION, "App key is not defined in AndroidManifest.xml");
        }
        return S2Api.requestPublishingEvents(appKey, rootEvent, leafEvents);
    }

    /**
     * 이 메소드는 ADID 정보를 가져와서 이번트에 채워준다. UI 스레드에서 실행할 경우 예외가 발생하기 때문에 항상 별도의 스레드에서 실행해야 한다.
     */
    private void fillAdidForEvent(final Event event) throws GooglePlayServicesNotAvailableException, GooglePlayServicesRepairableException, IOException {
        AdvertisingIdClient.Info idInfo = AdvertisingIdClient.getAdvertisingIdInfo(context);
        if (idInfo == null) {
            throw new IllegalStateException("Failed to get ADID from the device.");
        }
        event.setFrom(idInfo.getId());
        event.setAdidEnabled(!idInfo.isLimitAdTrackingEnabled());
    }


    /**
     * 애플리케이션이 foreground/background 전활될 때마다 타이머를 설정하고 해제하는 메소드들.
     */
    void addTimer() {
        eventsDataSource.open();
        TimerTask task = new TimerTask() {
            @Override
            public void run() {
                publishEvents();
            }
        };
        timer = new Timer(true);
        timer.scheduleAtFixedRate(task, 0, mBatchTime * 1000);
    }

    void removeTimer() {
        timer.cancel();
        timer = null;
        eventsDataSource.close();
    }


    public static class AlreadyInitializedException extends RuntimeException {
        private static final long serialVersionUID = 1L;
    }
}
